@@ -1,7 +1,9 @@
-from __future__ import unicode_literals
-from prompt_toolkit.filters import Condition, Never, Always, Filter, to_filter
+from __future__ import annotations
+
 import pytest
 
+from prompt_toolkit.filters import Always, Condition, Filter, Never, to_filter
+
 
 def test_never():
     assert not Never()()
@@ -13,7 +15,7 @@ def test_always():
 
 def test_invert():
     assert not (~Always())()
-    assert (~Never()())
+    assert ~Never()()
 
     c = ~Condition(lambda: False)
     assert c()
